/**
 ** dtmExecTaskCodels.c
 **
 ** Codels appeles par la tache dtmExecTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#include <portLib.h>
#include <float.h>
#include <limits.h>

#include <assert.h>
#include <stdlib.h>
#include <h2timeLib.h>

#include "server/dtmHeader.h"
#include <pomStruct.h>
#include <stereopixelStruct.h>

#include <viamStruct.h>

#include <libdtm.h>
#ifdef GDHE
#include <gdhe/GDHE.h>
#include <libdtm_gdhe.h>
#endif

/* Internal dtm structures: mntTrajectory is the one which will be
   used to fill the P3D dedicated poster, mntItinerary is the one used
   to plan explorations and/or long term itineraries. local_* are
   required for the fusion process to cope with dynamic elements of
   the scene. */
static DTM*	mntTrajectory;
static DTM*	local_mntTrajectory = NULL;
static bool	buildMntTrajectory = false;
static DtmInitParams	mntTrajectoryParams;

static DTM*	mntItinerary;
static DTM*	local_mntItinerary = NULL;
static bool	buildMntItinerary = false;
static POSTER_ID posterItinerary;
static DtmInitParams	mntItineraryParams;

/* Internal variables, just used in the functions defined here */
static DATA_IM3D	im3dStr; /* built upon stereopixel poster  */
static int im3dStrWidth, im3dStrHeight;
static int videoWidth, videoHeight;
static int		im3dStrState = 0; 

static DATA_IM3D	im3dVyn; /* Bult upon velodyne poster */
static int im3dVynWidth, im3dVynHeight;
static int		im3dVynState = 0; 

/* Required posters */
static POSTER_ID	cameraPosterId;
static ViamImageBank *adr_cameraPoster;	

static POSTER_ID	stereopixelPosterId;
static Spix3DImage*	adr_stereopixelPoster;

static POSTER_ID	velodynePosterId;
static velodyne3DImage*	adr_velodynePoster;

static POSTER_ID pomPosterId;

T3D		mainToOrigin;	/* Position of the data that are fused --
				   memorized to be able to apply the JumpDtm
				   function  */


/* Just to know the time spent in the functions */
/* static int		timeStart, timeEnd; */
/* static int		chronometer = 0; */

/* Produced posters */
static DTM_P3D_POSTER*	dtmP3dPoster;
static DTM_LABEL_POSTER*	dtmLabelPoster;

/* What follows is for the display */
static int		gdheConnection = 0; /* Indicates wheter gdhe is
					       connected or not */ 
static char		gdheCmdBuf[1024];

/* Some fcts prototypes */
extern void dtm_fillIm3dFromSpix3DImage (Spix3DImage* adr_stereopixelPoster, 
					 DATA_IM3D* im3d);
extern void dtm_fillIm3dFromVyn3DImage (velodyne3DImage* adr_velodynePoster, 
					DATA_IM3D* im3d);
extern int dtm_to_p3d_poster ( DTM* dtm, double p3dNbLines, double p3dNbCols,
		double zOffset, double zScale, T3D* mainToOrigin, DTM_P3D_POSTER*
		dtm_p3d_poster );
extern int write_dtm_p3d_poster ( DTM_P3D_POSTER* dtm, FILE* fichier );
extern int dtm_to_label_poster ( DTM* dtm, DTM_LABEL_POSTER* dtm_label_poster );

static
void set_strategy(DTM* mnt, DtmStrategy strategy)
{
	dtm_dynamic_fuse_strategy strat;
	switch (strategy.kind) {
		case DTM_FIXED_TRESHOLD:
			strat.strat = FIXED_TRESHOLD;
			break;
		case DTM_MEAN_TRESHOLD:
			strat.strat = MEAN_TRESHOLD;
			break;
		default:
			assert(false);
	}
	strat.treshold = strategy.treshold;
	dtm_dynamic_set_strategy(mnt, strat);
}

static
STATUS initialize_internal_dtm(DtmDestination destination, int* report)
{

  /* DTM structure allocation */
  if (destination == DTM_FOR_PATH_PLANNING) {
    if (mntTrajectory)
      destroy_dtm(mntTrajectory);
    mntTrajectory = create_dtm ();
    if (mntTrajectory == NULL) {
      fprintf(stderr, "dtm: can not create internal data structure");
      *report = S_dtm_INTERNAL_DTM_CREATION_FAILED;
      return FALSE;
    }
    
    /* DTM cell array allocation */
    if (init_dtm_cells (mntTrajectory, 
			mntTrajectoryParams.xScale, mntTrajectoryParams.yScale,
			mntTrajectoryParams.xNb, mntTrajectoryParams.yNb,
			mntTrajectoryParams.xOrigin, mntTrajectoryParams.yOrigin,
			mntTrajectoryParams.slidingMode,
			X_LIGNES) != DTM_OK) {
      /* NB: The only way of storage implemented is X_LIGNES = X are Rows */
      *report = S_dtm_INTERNAL_DTM_INIT_FAILED;
      destroy_dtm (mntTrajectory);
      mntTrajectory = NULL;
      return FALSE;
    }
  }

  if (destination == DTM_FOR_ITINERARY_PLANNING) {
    if (mntItinerary)
      destroy_dtm(mntItinerary);
    mntItinerary = create_dtm ();
    if (mntItinerary == NULL) {
      fprintf(stderr, "dtm: can not create internal data structure");
      *report = S_dtm_INTERNAL_DTM_CREATION_FAILED;
      return FALSE;
    }
    if (init_dtm_cells (mntItinerary, 
			mntItineraryParams.xScale, mntItineraryParams.yScale,
			mntItineraryParams.xNb, mntItineraryParams.yNb,
			mntItineraryParams.xOrigin, mntItineraryParams.yOrigin,
			mntItineraryParams.slidingMode,
			X_LIGNES) != DTM_OK) {
      /* NB: The only way of storage implemented is X_LIGNES = X are Rows */
      *report = S_dtm_INTERNAL_DTM_INIT_FAILED;
      destroy_dtm (mntItinerary);
      mntItinerary = NULL;
      return FALSE;
    }
  }

  return TRUE;
}

/*------------------------------------------------------------------------
 *
 * dtmExecTaskInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: initializes all the params with default values (executed at
 * the module initialisation - dtmTaskInit)
 * 
 * Retourne:    OK ou ERROR
 */

STATUS dtmExecTaskInit(int *report)
{
  STATUS status = OK;

  /* Initialisation params */
  SDI_F->initParams.xScale = DTM_DEFAULT_X_SCALE;
  SDI_F->initParams.yScale = DTM_DEFAULT_Y_SCALE;
  SDI_F->initParams.xNb = DTM_DEFAULT_X_NB;
  SDI_F->initParams.yNb = DTM_DEFAULT_Y_NB;
  SDI_F->initParams.xOrigin = DTM_DEFAULT_X_ORIGIN;
  SDI_F->initParams.yOrigin = DTM_DEFAULT_Y_ORIGIN;

  /* Connection params */
  SDI_F->connectionParams.stereoConnection = 0;
  strncpy (SDI_F->connectionParams.cameraPoster, DTM_DEFAULT_CAMERA_POSTER,
	   DTM_MAX_STRLEN);
  strncpy (SDI_F->connectionParams.stereoPoster, DTM_DEFAULT_STEREO_POSTER,
	   DTM_MAX_STRLEN);
  SDI_F->connectionParams.velodyneConnection = 0;
  strncpy (SDI_F->connectionParams.velodynePoster, DTM_DEFAULT_VELODYNE_POSTER,
	   DTM_MAX_STRLEN);
  SDI_F->connectionParams.sickConnection = 0;
  strncpy (SDI_F->connectionParams.sickPoster, DTM_DEFAULT_SICK_POSTER,
	   DTM_MAX_STRLEN);

  SDI_F->connected = 0;

  /* Fusion params */
  SDI_F->fusionParams.maxDistance = DTM_DEFAULT_MAX_DIST;
  SDI_F->fusionParams.video = DTM_DEFAULT_VIDEO;
  SDI_F->fusionParams.poseRefinement = DTM_DEFAULT_REFINEMENT;
  SDI_F->fusionParams.referenceFrame = DTM_DEFAULT_REFERENCE_FRAME;
  SDI_F->fusionParams.strategy.kind = DTM_DEFAULT_STRATEGY_KIND;
  SDI_F->fusionParams.strategy.treshold = DTM_DEFAULT_STRATEGY_TRESHOLD;

  /* Save params */
  strncpy (SDI_F->saveParams.baseName, DTM_DEFAULT_BASENAME,
	   DTM_MAX_STRLEN);
  SDI_F->saveParams.num = 0;
  SDI_F->saveParams.saveLocalDtm = DTM_DEFAULT_SAVELOCALDTM;
  SDI_F->saveParams.fileFormat = DTM_ASCII;

  /* Exported poster params */
  SDI_F->p3dPosterParams.nbLines = DTM_DEFAULT_POSTER_NB_LINES;
  SDI_F->p3dPosterParams.nbCols = DTM_DEFAULT_POSTER_NB_COLS;
  SDI_F->p3dPosterParams.zScale = DTM_DEFAULT_Z_SCALE;
  SDI_F->p3dPosterParams.zOffset = DTM_DEFAULT_Z_OFFSET;

  /* Display params */
  strncpy (SDI_F->displayParams.serverName, DTM_DEFAULT_DPY_SERVER,
	   DTM_MAX_STRLEN);
  strncpy (SDI_F->displayParams.displayFileName, DTM_DEFAULT_DPY_FILE,
	   DTM_MAX_STRLEN);
  SDI_F->displayParams.displayFileNumber = DTM_DEFAULT_DPY_NUMBER;
  SDI_F->displayParams.zOffset = DTM_DEFAULT_DPY_Z_OFFSET;
  SDI_F->displayParams.zScale = DTM_DEFAULT_DPY_Z_SCALE;
  SDI_F->displayParams.dpyColor = DTM_DEFAULT_DPY_COLOR;
  SDI_F->displayParams.dpyCrop = DTM_DEFAULT_DPY_CROP;
  SDI_F->displayParams.dpyLight = DTM_DEFAULT_DPY_LIGHT;

  /* Private structures */
  dtmP3dPoster = posterAddr (DTM_P3DPOSTER_POSTER_ID);
  if (dtmP3dPoster == NULL)
    return ERROR;

  return status;
}

/*------------------------------------------------------------------------
 *
 * dtmExecTaskEnd  --  Codel de terminaison
 *
 * Description: just free the internal structures. 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS dtmExecTaskEnd(void)
{
  if (mntTrajectory != NULL) {
    destroy_dtm (mntTrajectory);
    mntTrajectory = NULL;
  }
  if (mntItinerary != NULL) {
    destroy_dtm (mntItinerary);
    mntTrajectory = NULL;
  }
  
  if (im3dStrState != 0) {
    empty_data_im3d (&im3dStr);
  }
  if (im3dVynState != 0) {
    empty_data_im3d (&im3dVyn);
  }
  return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description: initializes the internals mnt 
 *
 * Reports:      OK 
 *              S_dtm_BAD_INIT_PARAMS
 *              S_dtm_INTERNAL_DTM_INIT_FAILED
 *              S_dtm_INTERNAL_DTM_CREATION_FAILED
 */

/* dtmInitExec  -  codel EXEC de Init
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmInitExec (DtmInitParams* initParams, int *report)
{
  STATUS e;
  unsigned long size;
#if 0
  int		iReduction, jReduction;	/* Reduction factor */
  double	rectifiedIntrinsinc[16]; /* Intrinsinc matrix */
#endif

  /* Check params */
  if (dtm_checkInitParams (initParams->xScale, initParams->yScale,
			   initParams->xNb, initParams->yNb,
			   X_LIGNES) != DTM_OK) {
    *report = S_dtm_BAD_INIT_PARAMS;
    return ETHER;
  }

  /* Check whether the DTM has already been initialized or not */
  if (initParams->destination == DTM_FOR_PATH_PLANNING) {
    memcpy (&mntTrajectoryParams, initParams, sizeof (DtmInitParams));
    buildMntTrajectory = true;
    if (mntTrajectory != NULL) {
      destroy_dtm (mntTrajectory);
      mntTrajectory = NULL;
    }
  }
  if (initParams->destination == DTM_FOR_ITINERARY_PLANNING) {
    memcpy (&mntItineraryParams, initParams, sizeof (DtmInitParams));
    buildMntItinerary = true;
    if (mntItinerary != NULL) {
      destroy_dtm (mntItinerary);
      mntItinerary = NULL;
    }

	// create the associated poster
	size = sizeof(struct DTM_LABEL_POSTER) + initParams->xNb * initParams->yNb;
	e = posterCreate("dtmLabelPoster", size, &posterItinerary);
	if (e) {
		destroy_dtm (mntItinerary);
		*report = S_dtm_POSTER_CREATION_FAILED;
		return ETHER;
	}
	dtmLabelPoster = posterAddr (posterItinerary);
	posterTake(posterItinerary, POSTER_WRITE);
	dtmLabelPoster->nbLines = initParams->xNb;
	dtmLabelPoster->nbCols = initParams->yNb;
	posterGive(posterItinerary);
  }
  
  if (!initialize_internal_dtm(initParams->destination, report))
      return ETHER;

  return ETHER;
}

/*------------------------------------------------------------------------
 * ReInit
 *
 * Description: Resets all the mnt cells to 0 (and also re-inits various other
 * field), so that 'mnt' is as if it had jsut been initialized.
 *
 * Reports:      OK
 *              S_dtm_INTERNAL_DTM_INIT_FAILED
 */

/* dtmReInitExec  -  codel EXEC of ReInit
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
dtmReInitExec(DtmResetParams* resetParams, int *report)
{
  initialize_internal_dtm(resetParams->destination, report);
  return ETHER;
}

static bool
open_and_check_camera_poster(const char* cameraPoster, int *report)
{
    if (SDI_F->fusionParams.video == DTM_USE_VIDEO) {
		/* Check the existence of the camera poster */
		if (posterFind(cameraPoster, &cameraPosterId) != OK) {
		  fprintf(stderr, "dtm: camera Poster %s not found\n", cameraPoster);
		  return false;
		} else {
		  fprintf(stderr, "dtm found camera poster %s\n", cameraPoster);
		}

		/* Get its address */
		adr_cameraPoster = posterAddr (cameraPosterId);
		if (adr_cameraPoster == NULL) {
		  *report = S_dtm_POSTER_NOT_FOUND;
		  return false;
		}

		if (posterTake(cameraPosterId, POSTER_READ) != OK) {
		  *report = errnoGet();
		  return false;
		}
		videoWidth = adr_cameraPoster->image[0].width;
		videoHeight = adr_cameraPoster->image[0].height;
		
		posterGive(cameraPosterId);

		if ((im3dStrWidth != videoWidth) || (im3dStrHeight != videoHeight)) {
		  *report = S_dtm_POSTERS_NOT_COMPATIBLE;
		  return false;
		}
	}

	return true;
}

/*------------------------------------------------------------------------
 * Connect
 *
 * Description: checks the imported poster existence and get their address,
 * initializes the internal mnt and the DATA_IM3D structure. 
 *
 * Reports:      OK 
 *              S_dtm_BAD_INIT_PARAMS
 *              S_dtm_INTERNAL_DTM_INIT_FAILED
 *              S_dtm_INTERNAL_DTM_CREATION_FAILED
 *              S_dtm_IM3D_CREATION
 *              S_dtm_POSTER_NOT_FOUND
 *		S_dtm_POSTERS_NOT_COMPATIBLE
 */

/* dtmInitExec  -  codel EXEC de Init
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmConnectExec (DtmConnectParams* connectionParams, int *report)
{
#if 0
  int		iReduction, jReduction;	/* Reduction factor */
  double	rectifiedIntrinsinc[16]; /* Intrinsinc matrix */
#endif

  /* Look up for POM Poster */
  if (posterFind(connectionParams->pomPoster, &pomPosterId) == ERROR) {
    fprintf(stderr, "dtm: cannot find pom poster\n");
    *report = S_dtm_POSTER_NOT_FOUND;
    return ETHER;
  }

  if (connectionParams->stereoConnection) {
    /* Check the existence of the stereopixel poster */
    if (posterFind (SDI_F->connectionParams.stereoPoster, &stereopixelPosterId) != OK) {
      fprintf (stderr, "dtm: cannot find poster %s\n", 
	       SDI_F->connectionParams.stereoPoster);
      *report = S_dtm_POSTER_NOT_FOUND;
      return ETHER;
    } else {
      fprintf(stderr, "dtm: found stereopixel poster %s\n", 
	      SDI_F->connectionParams.stereoPoster); 
    }
    /* Get its address */
    adr_stereopixelPoster = posterAddr (stereopixelPosterId);
    if (adr_stereopixelPoster == NULL) {
      *report = S_dtm_POSTER_NOT_FOUND;
      return ETHER;
    }

	if (posterTake (stereopixelPosterId, POSTER_READ) != OK) { 
	  *report = errnoGet(); 
	  return ETHER; 
	} 
	im3dStrWidth = adr_stereopixelPoster->width;
	im3dStrHeight = adr_stereopixelPoster->height;
	posterGive (stereopixelPosterId);

	if (!open_and_check_camera_poster(connectionParams->cameraPoster, report))
		return ETHER;

    /* Inits the im3dStr */
    if (im3dStrState != 0) {
      empty_data_im3d (&im3dStr);
      im3dStrState = 0;
    }
    if (im3dStrState == 0) {
      if (init_data_im3d (&im3dStr, im3dStrHeight, im3dStrWidth, 0, 0, 0) 
	  != ERR_IM3D_OK) {
	*report = S_dtm_IM3D_CREATION;
	return ETHER;
      }
      im3dStrState = 1;
    }
    im3dStr.header.capteur = CAPTEUR_STEREO;
    im3dStr.header.coordonnees = COORDONNEES_CARTESIENNES;
    fprintf(stderr, "dtm: uses %dx%d 3D images from stereopixel \n", 
	    im3dStrWidth, im3dStrHeight);

    SDI_F->connected = 1;
  }

  if (connectionParams->velodyneConnection) {
    /* Check the existence of the velodyne poster */
    if (posterFind(SDI_F->connectionParams.velodynePoster, &velodynePosterId) != OK) {
      fprintf(stderr, "dtm: velodyne Poster %s not found\n",
	      SDI_F->connectionParams.velodynePoster);
      return ETHER;
    } else {
      fprintf(stderr,
	      "dtm: found velodyne poster %s\n", 
	      SDI_F->connectionParams.velodynePoster);
    }
    
    /* Get its address */
    adr_velodynePoster = posterAddr (velodynePosterId);
    if (adr_velodynePoster == NULL) {
      *report = S_dtm_POSTER_NOT_FOUND;
      return ETHER;
    }
    
    im3dVynHeight = adr_velodynePoster->height;
    im3dVynWidth = adr_velodynePoster->maxScanWidth;
    
    /* Inits the im3dVyn */
    if (im3dVynState != 0) {
      empty_data_im3d (&im3dVyn);
      im3dVynState = 0;
    }
    if (im3dVynState == 0) {
      if (init_data_im3d (&im3dVyn, im3dVynHeight, im3dVynWidth, 0, 0, 0) 
	  != ERR_IM3D_OK) {
	*report = S_dtm_IM3D_CREATION;
	return ETHER;
      }
      im3dVynState = 1;
    }
    im3dVyn.header.capteur = CAPTEUR_LASER1;
    im3dVyn.header.coordonnees = COORDONNEES_CARTESIENNES;
    fprintf(stderr, "dtm: uses %dx%d 3D images from velodyne\n", 
	    im3dVynWidth, im3dVynHeight);

    SDI_F->connected = 1;
  }

  if (connectionParams->sickConnection) {
    *report = S_dtm_FONCTION_NOT_IMPLEMENTED;
    fprintf (stderr, "dtm: the module is currently not able to get Sick data\n");
    return ETHER;
  }

  return ETHER;
}


/*------------------------------------------------------------------------
 * FuseDtm
 *
 * Description:
 *
 * Reports:      OK 
 *              S_dtm_IM3D_SIZE_CHANGED
 *              S_dtm_POSTER_NOT_FOUND
 *              S_dtm_STEREOPIXEL_POSTER_CHANGED
 *              S_dtm_CAMERA_POSTER_CHANGED
 *              S_dtm_BAD_T3D_TYPE
 */

/* dtmFuseDtmExec  -  codel EXEC de FuseDtm
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmFuseDtmExec (int *report)
{
  int error;

  T3D		sensorToOrigin;

  double mat_intrinsec_rectified[9];

  if (!SDI_F->connected) {
    fprintf (stderr, "dtm: not connected to any 3D data module\n");
    *report = S_dtm_NOT_CONNECTED;
    return ETHER;
  }

  /* Fuse stereovision data */
  if (SDI_F->connectionParams.stereoConnection) {
    T3D		sensorToMainLeft;
    T3D		sensorToMainRight;

    t3dInit(&sensorToMainLeft, T3D_BRYAN, T3D_ALLOW_CONVERSION);
    t3dInit(&sensorToMainRight, T3D_BRYAN, T3D_ALLOW_CONVERSION);
    t3dInit(&mainToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
    t3dInit(&sensorToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
    
    /* The T3D below will be required when the pos refinement feature will be
     set up again */
    /*   T3DEuler      origin_2_CorrectedOrigin; */
    /*   T3DEuler      mainToOriginCorrected; */
    
    /* Get the stereopixel poster, fills the internal DATA_IM3D and TR3D with it
     */ 
    if (posterTake (stereopixelPosterId, POSTER_READ) != OK) { 
      *report = errnoGet(); 
      return ETHER; 
    } 
    if ((adr_stereopixelPoster->width != im3dStrWidth)
	|| (adr_stereopixelPoster->height != im3dStrHeight)) {
      posterGive (stereopixelPosterId);
      *report = S_dtm_IM3D_SIZE_CHANGED;
      return ETHER;
    }
    dtm_fillIm3dFromSpix3DImage (adr_stereopixelPoster, &im3dStr); 
    
    /* Now get the relevant T3D from the PomSensorPos of the stereopixel
       poster. CAREFUL: This assumes POM_EULER and euler from T3DEuler are both
       6 double arrays containing: yaw,pitch,roll,x,y,z */ 
    memcpy(&sensorToMainLeft.euler.euler, &(adr_stereopixelPoster->imageLeftPos.sensorToMain.euler), sizeof(POM_EULER));
    memcpy(&sensorToMainRight.euler.euler, &(adr_stereopixelPoster->imageRightPos.sensorToMain.euler), sizeof(POM_EULER));
    memcpy(&mainToOrigin.euler.euler, &(adr_stereopixelPoster->imageLeftPos.mainToOrigin.euler), sizeof(POM_EULER));
    memcpy(mat_intrinsec_rectified, adr_stereopixelPoster->intrinseque_rectified, sizeof(mat_intrinsec_rectified));
    posterGive (stereopixelPosterId);
    
    /* Compose the T3Ds to obtain SensorToMain, threshold the im3dStr depth,
     change the frame of its coordinates. */
    if (seuille_im3d (&im3dStr, 0, 0,
		      0, 0, 1, SDI_F->fusionParams.maxDistance)  != ERR_IM3D_OK) {
      fprintf(stderr, "dtm: error in thresholding the 3D image");
      *report = S_dtm_CALIFE_ERROR;
      return ETHER;
    }
    switch (SDI_F->fusionParams.referenceFrame) {
    case DTM_REFERENCE_FRAME_GLOBAL:
      /* Prepare the T3D for future conversions/compositions */
      sensorToMainLeft.flags |= T3D_ALLOW_CONVERSION;
      mainToOrigin.flags |= T3D_ALLOW_CONVERSION;
      t3dSetOk(&sensorToMainLeft);
      t3dSetOk(&mainToOrigin);
      t3dSetOk(&sensorToOrigin);
      t3dCompIn(&sensorToOrigin, &sensorToMainLeft, &mainToOrigin);
      error = change_repere_data_im3d (&im3dStr, &sensorToOrigin);
      break;
    case DTM_REFERENCE_FRAME_ROBOT:
      t3dCopyIn(&sensorToOrigin, &sensorToMainLeft);
      error = change_repere_data_im3d (&im3dStr, &sensorToOrigin);
      break;
    }
    if (error != ERR_IM3D_OK) {
      fprintf(stderr, "dtm: error in changing the 3D image frame");
      *report = S_dtm_CALIFE_ERROR;
      return ETHER;
    }
    
    if (buildMntTrajectory == true) {
      if (local_mntTrajectory == NULL) {
	// special crafted to compute initial vision
	sensorToMainLeft.euler.euler[0] = 0.0;
	sensorToMainRight.euler.euler[0] = 0.0;
	
	local_mntTrajectory = 
	  prepare_dynamic_fuse(&sensorToMainLeft, &sensorToMainRight,
			       mat_intrinsec_rectified,
			       im3dStrHeight, im3dStrWidth, 1,
			       SDI_F->fusionParams.maxDistance,
			       mntTrajectoryParams.xScale, 
			       mntTrajectoryParams.yScale);
	
	set_strategy(local_mntTrajectory, SDI_F->fusionParams.strategy);
      }
    }
    if (buildMntItinerary == true) {
      if (local_mntItinerary == NULL) {
	// special crafted to compute initial vision
	sensorToMainLeft.euler.euler[0] = 0.0;
	sensorToMainRight.euler.euler[0] = 0.0;
	
	local_mntItinerary = 
	  prepare_dynamic_fuse(&sensorToMainLeft, &sensorToMainRight,
			       mat_intrinsec_rectified,
			       im3dStrHeight, im3dStrWidth, 1,
			       SDI_F->fusionParams.maxDistance,
			       mntItineraryParams.xScale, 
			       mntItineraryParams.yScale);
	
	set_strategy(local_mntItinerary, SDI_F->fusionParams.strategy);
      }
    }

    /* Refine image position */
    /*   if (SDI_F->fusionParams.poseRefinement == DTM_USE_POSE_REFINEMENT) { */
    /*     origin_2_CorrectedOrigin = dtmPoseRefinement (&im3dStr, mnt,  */
    /* 						  SDI_F->fusionParams.refinementRate); */
    /*     if (origin_2_CorrectedOrigin != NULL) { */
    /* Put the im3dStr in the corrected frame */
    /*       error = change_repere_data_im3d (&im3dStr, origin_2_CorrectedOrigin); */
    /*       if (error != DTM_OK) { */
    /* 	cf_perror ("dtm: error in changing the 3D image frame"); */
    /* 	handle_error (NULL , NULL); */
    /* 	*report = S_dtm_CALIFE_ERROR; */
    /* 	return ETHER; */
    /*       } */
    /* Compute the 'MainToOriginCorrected' transfo */
    /*       t3dComp (&mainToOrigin, origin_2_CorrectedOrigin, &mainToOriginCorrected); */
    /*     } */
    /*     else { */
    /* No refinement produced */
    /*       t3_init (&MainToOriginCorrected); */
    /*       copy_tr3d (&MainToOrigin, &MainToOriginCorrected); */
    /*     } */
    /*   } */
    
    /* Fuse the image in the dtm */
    if (SDI_F->fusionParams.video == DTM_USE_VIDEO) {
      /* get the camera poster */
      if (posterTake (cameraPosterId, POSTER_READ) != OK) {
	/* TODO : le poster peut avoir disparu... Retenter un posterFind */
	*report = errnoGet ();
	return ETHER;
      }
      else {
	/* TODO : faire une copie de l'image pour ne pas monopoliser le poster
	 */ 
	if (buildMntTrajectory == true) {
	  if (dtm_fuseData (&im3dStr, &sensorToOrigin, mntTrajectory, DTM_AVERAGE, 1) != DTM_OK) {
	    posterGive (cameraPosterId);
	    *report = S_dtm_FUSION_ERROR;
	    return ETHER;
	  }
	  posterGive (cameraPosterId);
	}
	if (buildMntItinerary == true) {
	  if (dtm_fuseData (&im3dStr, &sensorToOrigin, mntItinerary, DTM_AVERAGE, 1) != DTM_OK) {
	    posterGive (cameraPosterId);
	    *report = S_dtm_FUSION_ERROR;
	    return ETHER;
	  }
	  posterGive (cameraPosterId);
	}
      }
    }
    else {
      if (buildMntTrajectory == true) {
	if (dtm_dynamic_fuse (&im3dStr, &mainToOrigin, &sensorToOrigin,
			      mntTrajectory, local_mntTrajectory, 0) != DTM_OK) {
	  *report = S_dtm_FUSION_ERROR;
	  return ETHER;
	}
      }
      if (buildMntItinerary == true) {
	if (dtm_dynamic_fuse (&im3dStr, &mainToOrigin, &sensorToOrigin,
			      mntItinerary, local_mntItinerary, 0) != DTM_OK) {
	  *report = S_dtm_FUSION_ERROR;
	  return ETHER;
	}      
      }
    }
  }

  /* Fuse velodyne data */
  if (SDI_F->connectionParams.velodyneConnection) {
    T3D		sensorToMain;

    t3dInit(&mainToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
    t3dInit(&sensorToMain, T3D_BRYAN, T3D_ALLOW_CONVERSION);
    t3dInit(&sensorToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
    
    if (posterTake (velodynePosterId, POSTER_READ) != OK) { 
      *report = errnoGet(); 
      return ETHER; 
    } 
    dtm_fillIm3dFromVyn3DImage (adr_velodynePoster, &im3dVyn);
    
    /* Now get the relevant T3D from the PomSensorPos of the velodyne poster */
    memcpy(&sensorToMain.euler.euler, 
	   &(adr_velodynePoster->position.sensorToMain.euler), sizeof(POM_EULER));
    memcpy(&mainToOrigin.euler.euler, 
	   &(adr_velodynePoster->position.mainToOrigin.euler), sizeof(POM_EULER));
	posterGive(velodynePosterId);
    
    /* Compose the T3Ds to obtain SensorToMain, threshold the im3dVyn depth,
     change the frame of its coordinates. */
/*     if (seuille_im3d (&im3dVyn, 1, SDI_F->fusionParams.maxDistance, */
/* 		      0, 0, 0, 0)  != ERR_IM3D_OK) { */
/*       fprintf(stderr, "dtm: error in thresholding the 3D image"); */
/*       *report = S_dtm_CALIFE_ERROR; */
/*       return ETHER; */
/*     } */
    switch (SDI_F->fusionParams.referenceFrame) {
    case DTM_REFERENCE_FRAME_GLOBAL:
      t3dCompIn(&sensorToOrigin, &sensorToMain, &mainToOrigin);
      break;
    case DTM_REFERENCE_FRAME_ROBOT:
      t3dCopyIn(&sensorToOrigin, &sensorToMain);
      break;
    default:
      break;
    }
    if (change_repere_data_im3d (&im3dVyn, &sensorToOrigin) != ERR_IM3D_OK) {
      fprintf(stderr, "dtm: error in changing the 3D image frame");
      *report = S_dtm_CALIFE_ERROR;
      return ETHER;
    }
    
    if (buildMntTrajectory == true) {
      if (local_mntTrajectory == NULL) {
	double dummyCameraCalib[9];
	// special crafted to compute initial vision
	sensorToMain.euler.euler[0] = 0.0;
	
	local_mntTrajectory = 
	  prepare_dynamic_fuse(&sensorToMain, &sensorToMain,
			       dummyCameraCalib,
			       0, 0, 1,
			       SDI_F->fusionParams.maxDistance,
			       mntTrajectoryParams.xScale,
			       mntTrajectoryParams.yScale);
	
	set_strategy(local_mntTrajectory, SDI_F->fusionParams.strategy);
      }
      
      if (dtm_dynamic_fuse (&im3dVyn, &mainToOrigin, &sensorToOrigin,
			    mntTrajectory, local_mntTrajectory, 0) != DTM_OK) {
	*report = S_dtm_FUSION_ERROR;
	return ETHER;
      }
    }
    if (buildMntItinerary == true) {
      if (local_mntItinerary == NULL) {
	double dummyCameraCalib[9];
	// special crafted to compute initial vision
	sensorToMain.euler.euler[0] = 0.0;
	
	local_mntItinerary = 
	  prepare_dynamic_fuse(&sensorToMain, &sensorToMain,
			       dummyCameraCalib,
			       0, 0, 1,
			       SDI_F->fusionParams.maxDistance,
			       mntItineraryParams.xScale,
			       mntItineraryParams.yScale);
	
	set_strategy(local_mntItinerary, SDI_F->fusionParams.strategy);
      }
      
      if (dtm_dynamic_fuse (&im3dVyn, &mainToOrigin, &sensorToOrigin,
			    mntItinerary, local_mntItinerary, 0) != DTM_OK) {
	*report = S_dtm_FUSION_ERROR;
	return ETHER;
      }
    }
  }
    
/*     if (dtm_fuseIm3dIma1 (&im3dVyn, &mainToOrigin, mntTrajectory, 0, NULL, 0, 0) != DTM_OK) { */
/*       *report = S_dtm_FUSION_ERROR; */
/*       return ETHER; */
/*     } */
/*   } */
  
  /* Fuse sick data */
  if (SDI_F->connectionParams.sickConnection) {
    *report = S_dtm_FONCTION_NOT_IMPLEMENTED;
    fprintf (stderr, 
	     "dtm: the module is currently not able to get Sick data\n");
    return ETHER;
  }

  return ETHER;
}

/*------------------------------------------------------------------------
 * JumpDtm
 *
 * Description: Account for a drastic change in the robot position
 *
 * Reports:      OK 
 */

/* dtmJumpDtmExec  -  codel EXEC de JumpDtm
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmJumpDtmExec (int *report)
{
  POM_POS newPos;
  T3D newPos_t3d;

  /* read the current robot position */
  if (dtmPOM_POSPosterRead(pomPosterId, &(newPos)) == ERROR) {
	  fprintf(stderr, "dtm: unable to read pomPos poster\n");
	  return ETHER;
  }

  t3dInit(&newPos_t3d, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  newPos_t3d.euler.euler[0] = newPos.mainToOrigin.euler.yaw;
  newPos_t3d.euler.euler[1] = newPos.mainToOrigin.euler.pitch;
  newPos_t3d.euler.euler[2] = newPos.mainToOrigin.euler.roll;
  newPos_t3d.euler.euler[3] = newPos.mainToOrigin.euler.x;
  newPos_t3d.euler.euler[4] = newPos.mainToOrigin.euler.y;
  newPos_t3d.euler.euler[5] = newPos.mainToOrigin.euler.z;

  /* Apply the jump with the current robot position and the last one that has
     been used to fuse the DTM */
  if (buildMntTrajectory == true)
    dtm_jumpDtm (mntTrajectory, &mainToOrigin, &newPos_t3d);
  if (buildMntItinerary == true)
    dtm_jumpDtm (mntItinerary, &mainToOrigin, &newPos_t3d);

  return ETHER;
}


/*------------------------------------------------------------------------
 * FillP3dPoster
 *
 * Description: 
 *
 * Reports:      OK 
 *              S_dtm_CALIFE_ERROR
 *              S_dtm_DTM_NOT_CREATED
 */

/* dtmFillP3dPosterExec  -  codel EXEC de FillP3dPoster
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmFillP3dPosterExec (int *report)
{
  int	error;
  int		event = ETHER;

  POM_POS pos;
  T3D mainToOrigin;

  if (buildMntTrajectory != true) {
    fprintf(stderr, "dtm: the internal dtm for P3D has not been created");
    *report = S_dtm_DTM_NOT_CREATED;
    return ETHER;
  }

  /* read robot position */
  if (dtmPOM_POSPosterRead(pomPosterId, &(pos)) == ERROR) {
	  fprintf(stderr, "dtm: unable to read pomPos poster\n");
	  return ETHER;
  }

  t3dInit(&mainToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  mainToOrigin.euler.euler[0] = pos.mainToOrigin.euler.yaw;
  mainToOrigin.euler.euler[1] = pos.mainToOrigin.euler.pitch;
  mainToOrigin.euler.euler[2] = pos.mainToOrigin.euler.roll;
  mainToOrigin.euler.euler[3] = pos.mainToOrigin.euler.x;
  mainToOrigin.euler.euler[4] = pos.mainToOrigin.euler.y;
  mainToOrigin.euler.euler[5] = pos.mainToOrigin.euler.z;

  posterTake (DTM_P3DPOSTER_POSTER_ID, POSTER_WRITE) ;
  error = dtm_to_p3d_poster (mntTrajectory, 
			     SDI_F->p3dPosterParams.nbLines,
			     SDI_F->p3dPosterParams.nbCols,
			     SDI_F->p3dPosterParams.zOffset,
			     SDI_F->p3dPosterParams.zScale,
			     &mainToOrigin, dtmP3dPoster); /* TODO : faire un memcpy ? */
  
  if (error != DTM_OK) {
    fprintf(stderr, "dtm: error in filling the dtm poster");
    *report = S_dtm_CALIFE_ERROR;
    posterGive (DTM_P3DPOSTER_POSTER_ID);
    return ETHER;
  }
  posterGive (DTM_P3DPOSTER_POSTER_ID);
  

  return event;
}

/*------------------------------------------------------------------------
 * FillLabelPoster
 *
 * Description: 
 *
 * Reports:      OK 
 *              S_dtm_CALIFE_ERROR
 *              S_dtm_DTM_NOT_CREATED
 */

/* dtmFillLabelPosterExec  -  codel EXEC de FillLabelPoster
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmFillLabelPosterExec (int *report)
{
  int	error;
  int		event = ETHER;

  if (buildMntItinerary != true) {
    fprintf(stderr, "dtm: the internal model of labels has not been created");
    *report = S_dtm_DTM_NOT_CREATED;
    return ETHER;
  }

  posterTake (posterItinerary, POSTER_WRITE) ;
  error = dtm_to_label_poster (mntItinerary, dtmLabelPoster); 
  posterGive (posterItinerary);

  if (error != DTM_OK) {
    fprintf(stderr, "dtm: error in filling the dtm poster");
    *report = S_dtm_CALIFE_ERROR;
    return ETHER;
  }

  return event;
}


#ifdef GDHE
/*------------------------------------------------------------------------
 * SaveGdheDtm
 *
 * Description: 
 *
 * Reports:      OK
 *              S_dtm_OPEN_FILE_FAILED
 */

/* dtmSaveGdheDtmExec  -  codel EXEC of SaveGdheDtm
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
dtmSaveGdheDtmExec(GEN_STRING256* gdheDtmFileName, int *report)
{
  FILE*		file;
  GDHE_MNT*	gdheDtm;

  DTM*	        mntCropped;

  /* Crop the internal dtm */
  mntCropped = dtm_autocrop (mntTrajectory);
  if (mntCropped == NULL) {
    *report = S_dtm_CREATION_CROPPED_MNT;
    return ETHER;
  }

  /* Converts the internal DTM into a GDHE_MNT structure */
  gdheDtm = dtm_createGdheMnt ();
  if (gdheDtm == NULL) {
    *report = S_dtm_CREATION_GDHE_MNT;
    return ETHER;
  }
  if (dtm_dtm2GdheMnt (mntCropped, DTM_DEFAULT_DPY_Z_SCALE,
		       DTM_DEFAULT_DPY_Z_OFFSET, DTM_DEFAULT_VIDEO, 
		       gdheDtm) != DTM_OK) {
    *report = S_dtm_CONVERSION_GDHE_MNT;
    return ETHER;
  }

  /* Open the file and write the gdheDtm */
  file = fopen(gdheDtmFileName->name, "w");
  if (file == NULL) {
    *report = S_dtm_OPEN_FILE_FAILED;
    return ETHER;
  }
  if (dtm_writeGdheMnt (gdheDtm, file) != DTM_OK) {
    *report = S_dtm_WRITE_DTM;
    return ETHER;
  }
  fclose (file);
  
  dtm_destroyGdheMnt (gdheDtm);
  destroy_dtm (mntCropped);

  return ETHER;
}
#endif


/*------------------------------------------------------------------------
 * SaveP3dPoster
 *
 * Description: just saves the DTM_P3D_POSTER in a file.
 *
 * Reports:      OK 
 *              S_dtm_NO_SUCH_FILE
 *              S_dtm_CALIFE_ERROR
 */

/* dtmSaveP3dPosterExec  -  codel EXEC de SaveP3dPoster
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmSaveP3dPosterExec (int *report)
{
  FILE*		file;
  int	error;
  char        fileName[DTM_MAX_STRLEN];
  
  /* Compose the file name */
  sprintf(fileName, "%s/dtm.%03d", 
	  SDI_F->saveParams.baseName, SDI_F->saveParams.num);
  
  file = fopen(fileName, "w");
  if (file == NULL) {
    fprintf(stderr, "dtm: cannot open file %s\n", fileName);
    *report = S_dtm_NO_SUCH_FILE;
    return ETHER;
  }
  
  (SDI_F->saveParams.num)++;
  
  posterTake (DTM_P3DPOSTER_POSTER_ID, POSTER_WRITE) ;
  error = write_dtm_p3d_poster (dtmP3dPoster, file);
  if (error != DTM_OK) {
    fprintf(stderr, "dtm: error in writing the dtm poster");
    *report = S_dtm_CALIFE_ERROR;
    posterGive (DTM_P3DPOSTER_POSTER_ID);
    fclose (file);
    return ETHER;
  }
  posterGive (DTM_P3DPOSTER_POSTER_ID);

  fclose (file);
  
  return ETHER;
}


/*------------------------------------------------------------------------
 * SaveInternalDtm
 *
 * Description: just saves the internal DTM structures in a file.
 *
 * Reports:      OK 
 *              S_dtm_NO_SUCH_FILE
 *              S_dtm_CALIFE_ERROR
 *
 *	TODO: it uses the same saveParams than the fct SaveP3dPoster, so take
 * care of the 'num' field increment. It is just to debug anyway.
 */

/* dtmSaveInternalDtmExec  -  codel EXEC de SaveInternalDtm
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmSaveInternalDtmExec (int *report)
{
  FILE*		file;
  char		fileName[DTM_MAX_STRLEN];
  
  if (buildMntTrajectory == true) {
    /* Compose the file name */
    sprintf(fileName, "%s/internalDtm.trajectory.%04d", 
	    SDI_F->saveParams.baseName, SDI_F->saveParams.num);
    
    file = fopen(fileName, "w");
    if (file == NULL) {
      fprintf(stderr, "dtm: cannot open %s\n", fileName);
      *report = S_dtm_NO_SUCH_FILE;
      return ETHER;
    }
    
    if(SDI_F->saveParams.fileFormat == DTM_BINARY) {
      if (dtm_writeBinaryDtm (mntTrajectory, file) != DTM_OK) {
	*report = S_dtm_WRITE_DTM;
	goto end;
      }
    }
    else {
      if (dtm_writeAsciiDtm (mntTrajectory, file) != DTM_OK) {
	*report = S_dtm_WRITE_DTM;
	goto end;
      }
    }
    fclose(file);
    file = NULL;
    
    if (SDI_F->saveParams.saveLocalDtm && (local_mntTrajectory != NULL)) {
      /* Compose the file name */
      sprintf(fileName, "%s/internalInnerDtm.trajectory.%04d", 
	      SDI_F->saveParams.baseName, SDI_F->saveParams.num);
      file = fopen(fileName, "w");
      if (file == NULL) {
	fprintf(stderr, "dtm: cannot open %s\n", fileName);
	*report = S_dtm_NO_SUCH_FILE;
	return ETHER;
      }
      
      if (SDI_F->saveParams.fileFormat == DTM_BINARY) {
	if (dtm_writeBinaryDtm (local_mntTrajectory, file) != DTM_OK) {
	  *report = S_dtm_WRITE_DTM;
	  goto end;
	}
      }
      else {
	if (dtm_writeAsciiDtm (local_mntTrajectory, file) != DTM_OK) {
	  *report = S_dtm_WRITE_DTM;
	  goto end;
	}
      }
    }
  }
  
  if (buildMntItinerary == true) {
    /* Compose the file name */
    sprintf(fileName, "%s/internalDtm.itinerary.%04d", 
	    SDI_F->saveParams.baseName, SDI_F->saveParams.num);
    
    file = fopen(fileName, "w");
    if (file == NULL) {
      fprintf(stderr, "dtm: cannot open %s\n", fileName);
      *report = S_dtm_NO_SUCH_FILE;
      return ETHER;
    }
    
    if(SDI_F->saveParams.fileFormat == DTM_BINARY) {
      if (dtm_writeBinaryDtm (mntItinerary, file) != DTM_OK) {
	*report = S_dtm_WRITE_DTM;
	goto end;
      }
    }
    else {
      if (dtm_writeAsciiDtm (mntItinerary, file) != DTM_OK) {
	*report = S_dtm_WRITE_DTM;
	goto end;
      }
    }
    fclose(file);
    file = NULL;
    
    if (SDI_F->saveParams.saveLocalDtm && (local_mntItinerary != NULL)) {
      /* Compose the file name */
      sprintf(fileName, "%s/internalInnerDtm.itineraries.%04d", 
	      SDI_F->saveParams.baseName, SDI_F->saveParams.num);
      file = fopen(fileName, "w");
      if (file == NULL) {
	fprintf(stderr, "dtm: cannot open %s\n", fileName);
	*report = S_dtm_NO_SUCH_FILE;
	return ETHER;
      }
      
      if (SDI_F->saveParams.fileFormat == DTM_BINARY) {
	if (dtm_writeBinaryDtm (local_mntItinerary, file) != DTM_OK) {
	  *report = S_dtm_WRITE_DTM;
	  goto end;
	}
      }
      else {
	if (dtm_writeAsciiDtm (local_mntItinerary, file) != DTM_OK) {
	  *report = S_dtm_WRITE_DTM;
	  goto end;
	}
      }
    }
  }
  
  
 end:
  (SDI_F->saveParams.num)++;
  if (file != NULL)
    fclose (file);
  return ETHER;
}


/*------------------------------------------------------------------------
 *
 * dtmDisplayTaskInit  --  Codel d'initialisation (SDI/f, ...) de la tâche de display
 *
 * Description: initializes the display params with default values (executed at
 * the module initialisation - dtmTaskInit)
 * 
 * Retourne:    OK
 */

STATUS dtmDisplayTaskInit(int *report)
{
  STATUS status = OK;

  /* Display initialisation and display params */
  strncpy (SDI_F->displayParams.serverName, DTM_DEFAULT_DPY_SERVER,
	   DTM_MAX_STRLEN);
  strncpy (SDI_F->displayParams.displayFileName, DTM_DEFAULT_DPY_FILE,
	   DTM_MAX_STRLEN);
  SDI_F->displayParams.displayFileNumber = DTM_DEFAULT_DPY_NUMBER;
  SDI_F->displayParams.zOffset = DTM_DEFAULT_DPY_Z_OFFSET;
  SDI_F->displayParams.zScale = DTM_DEFAULT_DPY_Z_SCALE;
  SDI_F->displayParams.dpyColor = DTM_DEFAULT_DPY_COLOR;
  SDI_F->displayParams.dpyColor = DTM_DEFAULT_DPY_COLOR;

  return status;
}

#ifdef GDHE
/*------------------------------------------------------------------------
 *
 * dtmDisplayTaskEnd  --  Codel de terminaison de la tâche de display
 *
 * Description: just disconnect (if already connected)
 * 
 * Retourne:    OK
 */

STATUS dtmDisplayTaskEnd(void)
{
/*   if (gdheConnection != 0) */
/*     disconnect (); */
  return OK;
}

/*------------------------------------------------------------------------
 * InitDisplay
 *
 * Description: Initializes a display. Opens a connection with gdhe, save the
 * full internal dtm in a file understandable by gdhe, and displays it.
 *
 * Reports:      OK 
 *              S_dtm_NO_CONNECTION_POSSIBLE
 *              S_dtm_OPEN_FILE_FAILED
 *		S_dtm_BAD_DISPLAY_PARAMS
 *              S_dtm_CALIFE_ERROR
 * 
 */

/* dtmInitDisplayExec  -  codel EXEC de InitDisplay
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmInitDisplayExec (DtmDisplayParams *displayParams, int *report)
{
  FILE*		file;
  GDHE_MNT*	gdheDtm;
  int	error;
  char	displayFileName[DTM_MAX_STRLEN];
	
  /* Params check */
  if ((displayParams->zScale <= 0) || (displayParams->zOffset < 0.0)) {
    *report = S_dtm_BAD_DISPLAY_PARAMS;
    return ETHER;
  }

  /* Connect with gdhe */
  if (gdheConnection == 0) {
    if (get_connection(SDI_F->displayParams.serverName) == FALSE) {
      *report = S_dtm_NO_CONNECTION_POSSIBLE;
      return ETHER;
    }
    gdheConnection = 1;
  }

  /* Allocate and fill the structure readable by gdhe */
  gdheDtm = dtm_createGdheMnt ();
  if (gdheDtm == NULL) {
    fprintf(stderr, "dtm: error in dtm_createGdheMnt");
    *report = S_dtm_CALIFE_ERROR;
    return ETHER;
  }
  error = dtm_dtm2GdheMnt (mntTrajectory, 
				 SDI_F->displayParams.zScale,
				 SDI_F->displayParams.zOffset,
				 SDI_F->displayParams.dpyColor,
				 gdheDtm);
  if (error != DTM_OK) {
    fprintf(stderr,"dtm: error in dtm_dtm2GdheMnt");
    *report = S_dtm_CALIFE_ERROR;
    return ETHER;
  }

  /* Open the file and write the gdheDtm */
  sprintf (displayFileName, "%s.%04d",
	   SDI_F->displayParams.displayFileName,
	   SDI_F->displayParams.displayFileNumber);
  file = fopen(displayFileName, "w");
  if (file == NULL) {
    fprintf(stderr, "dtm: cannot open %s\n", displayFileName);
    *report = S_dtm_OPEN_FILE_FAILED;
    return ETHER;
  }
  error = dtm_writeGdheMnt (gdheDtm, file);
  if (error != DTM_OK) {
    fprintf(stderr, "dtm: error in dtm_writeGdheMnt");
    *report = S_dtm_CALIFE_ERROR;
    return ETHER;
  }
  SDI_F->displayParams.displayFileNumber += 1;
  fclose (file);
  
  dtm_destroyGdheMnt (gdheDtm);
  
  /* Now sends the commands to gdhe */
  sprintf (gdheCmdBuf, "initMntDisplay %s dtmModel %d %d %d",
	   displayFileName, 
	   SDI_F->displayParams.dpyColor, 
	   SDI_F->displayParams.dpyCrop,
	   SDI_F->displayParams.dpyLight);
  eval_expression (gdheCmdBuf);

  return ETHER;
}

ACTIVITY_EVENT dtmStopDisplayExec (int *report)
{
  /* Connect with gdhe */
  if (gdheConnection != 0)
    disconnect ();

  gdheConnection = 0;
  
  return ETHER;
}

/*------------------------------------------------------------------------
 * UpdateDisplay
 *
 * Description: updates a display
 *
 * Reports:      OK 
 *              S_dtm_NOT_CONNECTED
 *              S_dtm_OPEN_FILE_FAILED
 *              S_dtm_CALIFE_ERROR
 *
 */

/* dtmUpdateDisplayExec  -  codel EXEC de UpdateDisplay 
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT dtmUpdateDisplayExec (int *report)
{
  FILE*		file;
  GDHE_MNT*	gdheDtm;
  int		error;
  char	displayFileName[DTM_MAX_STRLEN];

  /* Check the connection with gdhe */
  if (gdheConnection == 0) {
    *report = S_dtm_NOT_CONNECTED;
    return ETHER;
  }

  /* Allocate and fill the structure readable by gdhe */
  gdheDtm = dtm_createGdheMnt ();
  if (gdheDtm == NULL) {
    fprintf(stderr, "dtm: error in dtm_createGdheMnt");
    *report = S_dtm_CALIFE_ERROR;
    return ETHER;
  }
  error = dtm_lastDtm2GdheMnt (mntTrajectory, 
			       SDI_F->displayParams.zScale,
			       SDI_F->displayParams.zOffset,
			       SDI_F->displayParams.dpyColor,
			       gdheDtm);
  if (error != DTM_OK) {
    if (error == ERR_DTM_NO_CELLS_UPDATED) {
      /* Nothing to do, no cell have been updated inthe dtm */
      /*       fprintf (stderr, "dtm: nothing to update in the display\n"); */
      return ETHER;
    }
    fprintf(stderr, "dtm: error in dtm_lastDtm2GdheMnt");
    *report = S_dtm_CALIFE_ERROR;
    return ETHER;
  }
  
  /* Open the file and write the gdheDtm */
  sprintf (displayFileName, "%s.chunk.%04d",
	   SDI_F->displayParams.displayFileName,
	   SDI_F->displayParams.displayFileNumber);
  file = fopen(displayFileName, "w");
  if (file == NULL) {
    fprintf(stderr, "dtm: cannot open %s\n", displayFileName);
    *report = S_dtm_OPEN_FILE_FAILED;
    return ETHER;
  }
  
  error = dtm_writeGdheMnt (gdheDtm, file);
  if (error != DTM_OK) {
    fprintf(stderr, "dtm: error in dtm_writeGdheMnt");
    *report = S_dtm_CALIFE_ERROR;
    return ETHER;
  }
  SDI_F->displayParams.displayFileNumber += 1;
  fclose (file);
  
  dtm_destroyGdheMnt (gdheDtm);
  
  /* Now sends the commands to gdhe */
  sprintf (gdheCmdBuf, "updateMnt dtmModel %s", displayFileName);
  eval_expression (gdheCmdBuf);

  return ETHER;
}

#endif


/*------------------------------------------------------------------------
 * dtmSetGeodesicOriginExec
 *
 * Description:    
 * Report (report): OK 
 *                 S_dtm_CALIFE_ERROR
 *
 * Return:    OK ou ERROR
 */

ACTIVITY_EVENT dtmSetGeodesicOriginExec (DtmGeodesicOrigin *geodesicOrigin, int *report)
{
  STATUS status = OK;
  
  if (buildMntTrajectory == true) {
    if (dtm_setUTMGeodesicReference (mntTrajectory, 0, 
				     geodesicOrigin->easting,
				     geodesicOrigin->northing,
				     geodesicOrigin->altitude,
				     0.0) != DTM_OK) {
      fprintf(stderr, "dtm: errors in the geodesic origin");
      *report = S_dtm_CALIFE_ERROR;
      status = ERROR;
    }
  }
  if (buildMntItinerary == true) {
    if (dtm_setUTMGeodesicReference (mntItinerary, 0, 
				     geodesicOrigin->easting,
				     geodesicOrigin->northing,
				     geodesicOrigin->altitude,
				     0.0) != DTM_OK) {
      fprintf(stderr, "dtm: errors in the geodesic origin");
      *report = S_dtm_CALIFE_ERROR;
      status = ERROR;
    }
  }
  
  return status;
}


/*------------------------------------------------------------------------
 * dtmSetFusionParamsCntrl  -  control codel of CONTROL request SetFusionParams
 *
 * Description: just checks the validity of the DTM fusion params.
 * Report (bilan): OK 
 *                 S_dtm_BAD_FUSION_PARAMS
 *
 * Return:    OK ou ERROR
 */

STATUS dtmSetFusionParamsCntrl (DtmFusionParams *fusionParams, int *bilan)
{
  STATUS status = OK;

  if (fusionParams->maxDistance <= 0.0) {
    fprintf(stderr, "maximum distance must be positive\n");
    *bilan = S_dtm_BAD_FUSION_PARAMS;
    status = ERROR;
  }
  if ((fusionParams->video != DTM_USE_VIDEO)
      && (fusionParams->video != DTM_NO_VIDEO)) {
    fprintf (stderr, "Bad param for 'fuseVideo'\n");
    *bilan = S_dtm_BAD_FUSION_PARAMS;
    status = ERROR;
  }

  if ((fusionParams->video == DTM_USE_VIDEO) &&
	  (SDI_F->connectionParams.stereoConnection)) {
	  if (!open_and_check_camera_poster(SDI_F->connectionParams.cameraPoster,
										bilan)) {
			status = ERROR;
	  }
  }

  if ((fusionParams->poseRefinement != DTM_USE_POSE_REFINEMENT)
      && (fusionParams->poseRefinement != DTM_NO_POSE_REFINEMENT)) {
    fprintf (stderr, "Bad param for 'poseRefinement'\n");
    *bilan = S_dtm_BAD_FUSION_PARAMS;
    status = ERROR;
  }
  if (fusionParams->refinementRate <= 0) {
    fprintf (stderr, "Refinement sub-sample rate distance must be positive\n");
    *bilan = S_dtm_BAD_FUSION_PARAMS;
    status = ERROR;
  }
  
  if (local_mntTrajectory != NULL) 
	  set_strategy(local_mntTrajectory, SDI_F->fusionParams.strategy);
  if (local_mntItinerary != NULL) 
	  set_strategy(local_mntItinerary, SDI_F->fusionParams.strategy);

  return status;
}

